@@ -3,7 +3,7 @@ package Stepford::Types::Internal;
 use strict;
 use warnings;
 
-our $VERSION = '0.003009';
+our $VERSION = '0.004000';
 
 use MooseX::Types::Common::String qw( NonEmptyStr );
 use MooseX::Types::Moose qw( ArrayRef Defined Str );
@@ -22,6 +22,8 @@ use MooseX::Types -declare => [
         )
 ];
 
+use namespace::clean;
+
 subtype PossibleClassName, as Str, inline_as {
     ## no critic (Subroutines::ProtectPrivateSubs)
     $_[0]->parent->_inline_check( $_[1] ) . ' && '
@@ -63,13 +65,19 @@ __END__
 
 =pod
 
+=encoding UTF-8
+
 =head1 NAME
 
 Stepford::Types::Internal - Internal type definitions for Stepford
 
 =head1 VERSION
 
-version 0.003009
+version 0.004000
+
+=head1 SUPPORT
+
+Bugs may be submitted through L<https://github.com/maxmind/Stepford/issues>.
 
 =head1 AUTHOR
 
@@ -77,7 +85,7 @@ Dave Rolsky <drolsky@maxmind.com>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2016 by MaxMind, Inc.
+This software is copyright (c) 2014 - 2017 by MaxMind, Inc.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
